# Algorithme MinMax et Alpha-Beta

# Objectif

Votre objectif sera d’implémenter l’algorithme MinMax et l’algorithme Alpha-Beta sur le jeu Puissance 4 . Vous partirez d’un code Java existant afin de vous focaliser essentiellement sur ces algorithmes. 

# Récupérer le code Java

Les sources sont sur [A.I.C.O.I](https://gvipers.imt-nord-europe.fr/franck.tempet/aicoi). Pour récupérer les sources taper la commande suivante :

```bash
git clone http://gvipers.imt-nord-europe.fr/franck.tempet/aicoi.git
```

## Quelques explications sur le code

### La classe Player

Cette classe contient la méthode **main** du programme. Elle crée un joueur «*humain*» ainsi qu’un joueur «*Ordinateur*». L’ordinateur joue les coups de manière aléatoire.   

### La classe Plateau

Regarder les méthodes contenues dans cette classe ainsi que les commentaires associés. Cette classe est une classe abstraite qui liste les méthodes nécessaires à la manipulation d’un jeu de plateau à deux joueurs. Vous aurez besoin essentiellement des méthodes suivantes :

> - getListeCoups()
> - joueCoup()
> - annuleDernierCoup()
> - partieTerminee() 
> - partieNulle() 

### La classe GrillePuissance4

Cette classe représente la grille du jeu Puissance 4. Cette classe hérite de Plateau, elle implémente donc toutes les méthodes abstraites présentes dans la classe Plateau. 

### La classe Evaluation 

Cette classe contient une méthode statique `public static int eval(Plateau _plateau, Joueur _joueur)` qui donne un score à un joueur passé en paramètre pour une position donnée (`_plateau`).

# Implémenter les algorithmes

Regardez en détail le contenu de la classe **Player**. La création de l’IA ( *joueurOrdi* ) nécessite  d’appeler la méthode **setAlgoRecherche** en lui passant en paramètre une instance d’une classe qui implémente l’interface **AlgoRecherche**. Dans notre cas c'est la classe **AlgoRechercheAleatoire** qui implémente cette interface. Elle implémente donc la méthode suivante :

```java
public Coup meilleurCoup(Plateau _plateau, Joueur _joueur)
```

On doit passer à cette méthode le plateau du jeu ( *ici GrillePuissance4* ) , le joueur à qui c'est le tour de jouer.

L'algorithme aléatoire récupère la liste des coups possibles à partir de la position courante et en choisit un au hasard.

```java
Random rnd;

public Coup meilleurCoup(Plateau _plateau, Joueur _joueur) {    

    ArrayList<Coup> coups = _plateau.getListeCoups(_joueur);   
    return coups.get(rnd.nextInt( coups.size()));
}
```

Vous devrez donc écrire une classe AlgoMinMax et une classe AlgoAlphaBeta. Ces deux classes devront implémenter également l’interface *AlgoRecherche* ce qui revient à écrire la méthode *meilleurCoup*.

# Tester son IA

La classe Player actuelle vous permet de jouer contre votre IA. Vous pouvez  modifier facilement cette classe afin de faire jouer deux IA l’une contre l’autre.

## Humain contre une IA qui joue de manière aléatoire

```java
public class Player {

    public static void main(String args[]) {

        JoueurHumain humain = new JoueurHumain("Humain");
        JoueurOrdi joueurOrdi = new JoueurOrdi("Ordi");

        AlgoRechercheAleatoire aleatoire = new AlgoRechercheAleatoire(); 
        joueurOrdi.setAlgoRecherche(aleatoire);

        GrillePuissance4 grille = new GrillePuissance4();

        Arbitre arbitre = new Arbitre(grille, humain, joueurOrdi);

        arbitre.startNewGame(true); // Demarre une partie en affichant la grille du jeu

    }

}
```

## Tournoi entre une IA aléatoire et une IA utilisant l'algorithme MinMax

```java
public class Player {

    public static void main(String args[]) {

        JoueurOrdi joueurAleatoire = new JoueurOrdi("Aleatoire");
        JoueurOrdi joueurMinMax = new JoueurOrdi("MinMax");

        AlgoRechercheAleatoire AlgoAleatoire = new AlgoRechercheAleatoire(); 
        joueurAleatoire.setAlgoRecherche(AlgoAleatoire);

        AlgoMinMax algoMinMax = new AlgoMinMax(joueurAleatoire, joueurMinMax); 
        joueurMinMax.setAlgoRecherche(algoMinMax);

        GrillePuissance4 grille = new GrillePuissance4();

        Arbitre arbitre = new Arbitre(grille, joueurAleatoire, joueurMinMax);
        arbitre.startTournament(100, false); // 100 parties sans afficher la grille du jeu.

    }

}
```


