package puissance4;

import java.util.ArrayList;

import jeuplateau.Plateau;
import jeuplateau.Joueur;
import jeuplateau.Jeton;
import jeuplateau.Coup;
import jeuplateau.Piece;

public class GrillePuissance4 extends Plateau {

    private static final int NB_COLONNE = 7;
    private static final int NB_LIGNE = 6;

    // Grille de jeu Puissance 4.
    Jeton[][] grille;

    // Nombre de jetons présents dans chaque colonne de la grille.
    int[] hauteurColonne;

    // Nombre de coups joués.
    int nbCoupJoue;

    Joueur vainqueur;

    boolean isPartieNulle, isPartieGagnee;

    // Historique des coups joués.
    CoupPuissance4[] historiqueCoups;

    // Variables pour la sauvegarde de la position actuelle

    // Grille qui est une copie de la grille actuelle. Utiliser pour sauvegarder la
    // position.
    Jeton[][] grilleSav;
    // Sauvegarde du nombre de coups joués.
    int nbCoupJoueSav;
    // Sauvegarde des hauteurs des colonnes.
    int[] hauteurColonneSav;

    // Ordre dans lequel le générateur de coup parcours les colonnes.
    int[] ordreScanneColonne = { 0, 1, 2, 3, 4, 5, 6 };

    public GrillePuissance4() {
        grille = new Jeton[NB_COLONNE][NB_LIGNE];
        grilleSav = new Jeton[NB_COLONNE][NB_LIGNE];
        hauteurColonne = new int[NB_COLONNE];
        hauteurColonneSav = new int[NB_COLONNE];
        historiqueCoups = new CoupPuissance4[NB_COLONNE * NB_LIGNE];

        init();
    }

    @Override
    public void init() {

        nbCoupJoue = 0;
        vainqueur = null;
        isPartieGagnee = false;
        isPartieNulle = false;

        for (int c = 0; c < this.getNbColonnes(); c++) {
            hauteurColonne[c] = 0;
            for (int l = 0; l < this.getNbLignes(); l++) {
                grille[c][l] = null;
                historiqueCoups[l * NB_COLONNE + c] = null;
            }
        }
    }

    @Override
    public Piece getPiece(int _x, int _y) {
        return grille[_x][_y];
    }

    @Override
    public void joueCoup(Coup _coup) {

        if (partieTerminee())
            return;

        CoupPuissance4 coup = (CoupPuissance4) _coup;

        grille[coup.getColonne()][hauteurColonne[coup.getColonne()]] = coup.getJeton();
        hauteurColonne[coup.getColonne()]++;

        historiqueCoups[nbCoupJoue] = coup;
        nbCoupJoue++;

        isPartieGagnee = partieGagnee();
        isPartieNulle = !isPartieGagnee && isGrillePleine();

    }

    @Override
    public void annuleDernierCoup() {
        nbCoupJoue--;

        CoupPuissance4 dernierCoup = historiqueCoups[nbCoupJoue];
        hauteurColonne[dernierCoup.getColonne()]--;

        grille[dernierCoup.getColonne()][hauteurColonne[dernierCoup.getColonne()]] = null;
        isPartieGagnee = false;
        isPartieNulle = false;
    }

    @Override
    public int getNbColonnes() {
        return NB_COLONNE;
    }

    @Override
    public int getNbLignes() {
        return NB_LIGNE;
    }

    @Override
    public boolean partieTerminee() {
        return isPartieGagnee || isPartieNulle;
    }

    private boolean isGrillePleine() {
        return nbCoupJoue == NB_COLONNE * NB_LIGNE;
    }

    @Override
    public boolean partieGagnee() {
        int[][][] dir = { { { 1, 0 }, { 1, 1 }, { 0, 1 }, { 1, -1 } }, // vers la droite
                { { -1, 0 }, { -1, -1 }, { 0, -1 }, { -1, 1 } } // vers la gauche
        };
        int x, y;

        int nbJetonAligne;

        if (nbCoupJoue < 7)
            return false;

        Joueur dernierJoueur = historiqueCoups[nbCoupJoue - 1].getJeton().getJoueur();

        /* Regarde si le dernier coup est gagnant */

        // Boucle sur les 4 directions
        for (int d = 0; d < 4; d++) {
            nbJetonAligne = 1;

            // Boucle sur les 2 sens.
            for (int s = 0; s < 2; s++) {

                x = historiqueCoups[nbCoupJoue - 1].getColonne() + dir[s][d][0];
                if (x >= this.getNbColonnes() || x < 0)
                    continue;
                y = hauteurColonne[historiqueCoups[nbCoupJoue - 1].getColonne()] - 1 + dir[s][d][1];

                while (x < this.getNbColonnes() && x >= 0 && y < this.getNbLignes() && y >= 0 && grille[x][y] != null
                        && grille[x][y].getJoueur() == dernierJoueur) {
                    nbJetonAligne++;

                    if (nbJetonAligne >= 4) {
                        vainqueur = dernierJoueur;
                        return true;
                    }
                    x += dir[s][d][0];
                    y += dir[s][d][1];
                }
            }
        }

        return false;
    }

    @Override
    public boolean partieNulle() {
        return isPartieNulle;
    }

    @Override
    public Joueur vainqueur() {
        return vainqueur;
    }

    @Override
    public ArrayList<Coup> getListeCoups(Joueur _joueur) {
        ArrayList<Coup> listeCoups = new ArrayList<Coup>();

        for (int c = 0; c < this.getNbColonnes(); c++) {
            if (hauteurColonne[ordreScanneColonne[c]] < NB_LIGNE)
                listeCoups.add(new CoupPuissance4(ordreScanneColonne[c], new Jeton(_joueur)));
        }

        return listeCoups;
    }

    @Override
    public boolean isValide(Coup _coup) {
        if (_coup == null)
            return false;
        CoupPuissance4 coup = (CoupPuissance4) _coup;
        if (coup.getColonne() < 0 || coup.getColonne() >= getNbColonnes())
            return false;
        return hauteurColonne[coup.getColonne()] < NB_LIGNE;
    }

    @Override
    public Coup stringToCoup(String _coup, Joueur _joueur) {
        try {
            int colonne = Integer.valueOf(_coup.charAt(0) + "");
            return new CoupPuissance4(colonne, new Jeton(_joueur));
        } catch (NumberFormatException e) {
            return null;
        } catch (StringIndexOutOfBoundsException e) {
            return null;
        }
    }

    @Override
    public void sauvegardePosition() {
        for (int c = 0; c < this.getNbColonnes(); c++)
            for (int l = 0; l < this.getNbLignes(); l++)
                grilleSav[c][l] = grille[c][l];

        nbCoupJoueSav = nbCoupJoue;
    }

    @Override
    public void restaurePosition() {
        for (int c = 0; c < this.getNbColonnes(); c++)
            for (int l = 0; l < this.getNbLignes(); l++)
                grille[c][l] = grilleSav[c][l];

        vainqueur = null;
        nbCoupJoue = nbCoupJoueSav;
        isPartieNulle = false;
        isPartieGagnee = false;
    }

    @Override
    public Coup getDernierCoup() {
        return historiqueCoups[nbCoupJoue - 1];
    }

}
